package com.tocea.easycoverage.framework;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.tocea.easycoverage.framework.api.IEasycoverageChecker;
import com.tocea.easycoverage.framework.api.IInstanceProvider;
import com.tocea.easycoverage.framework.exception.EasycoverageAssertionError;
import com.tocea.easycoverage.framework.providers.DefaultInstanceProvider;
import com.tocea.easycoverage.framework.utils.Utils;

/**
 * 
 * A helper class to run checks on to build other runners on top of it. It takes
 * an instance provider, a list of classes, a list of class checkers and a list
 * of method checkers. getCheckers returns a set of class or method checkers
 * (depending on the parameter), while runChecks will run all the possible
 * checks and print the results in System.out.
 * 
 * @author Romain Delamare
 * 
 */
public class CheckersRunner {

	private static final Logger					LOGGER	= LoggerFactory
																.getLogger(CheckersRunner.class);

	private IInstanceProvider					provider;
	private Set<Class<?>>						classes;
	private Set<Class<? extends ClassChecker>>	classCheckers;
	private Set<Class<? extends MethodChecker>>	methodCheckers;
	private int									timeout	= 0;

	// TODO clean and simplify;
	/**
	 * <p>
	 * Runs checks on all the classes previously specified. For each class it
	 * will run all the class checkers that can be applied, and for each method
	 * it will run all the method checkers that can be applied.
	 * </p>
	 * 
	 * <p>
	 * Prints the result in System.out.
	 * </p>
	 */
	public void runChecks() {
		final PrintStream out = System.out;
		System.setOut(new PrintStream(new ByteArrayOutputStream()));
		int runChecks = 0;
		int failedChecks = 0;
		final ExecutorService executor = Executors.newCachedThreadPool();
		for (final Class<?> clazz : getClasses()) {
			out.println("****** " + clazz.getName() + " ******");
			final Set<ClassChecker> classCheckers = getCheckers(clazz);
			runChecks += classCheckers.size();
			for (final ClassChecker checker : classCheckers) {
				final Future<?> future = executor.submit(new CheckerTask(
						checker));
				try {
					if (timeout > 0)
						future.get(timeout, TimeUnit.MILLISECONDS);
					else
						future.get();
				} catch (final InterruptedException e) {
					failedChecks++;
					out.println(" - " + checker.getClass().getSimpleName()
							+ " failed: interrupted");
				} catch (final TimeoutException e) {
					if (future != null && !future.isDone()
							&& !future.isCancelled())
						future.cancel(true);
					failedChecks++;
					out.println(" - " + checker.getClass().getSimpleName()
							+ " failed: timeout");
				} catch (final ExecutionException e) {
					failedChecks++;
					out.println(" - " + checker.getClass().getSimpleName()
							+ " failed: " + e.getMessage());
				}
			}
			for (final Method method : Utils.getDeclaredPublicMethods(clazz)) {
				final Set<MethodChecker> methodCheckers = getCheckers(method);
				if (methodCheckers.isEmpty())
					continue;
				out.println("\t****** " + Utils.getQualifiedName(method)
						+ " ******");
				runChecks += methodCheckers.size();
				for (final MethodChecker checker : methodCheckers) {
					final Future<?> future = executor.submit(new CheckerTask(
							checker));
					try {
						if (timeout > 0)
							future.get(timeout, TimeUnit.MILLISECONDS);
						else
							future.get();
					} catch (final InterruptedException e) {
						failedChecks++;
						out.println("\t - "
								+ checker.getClass().getSimpleName()
								+ " failed: interrupted");
					} catch (final TimeoutException e) {
						if (future != null && !future.isDone()
								&& !future.isCancelled())
							future.cancel(true);
						failedChecks++;
						out.println("\t - "
								+ checker.getClass().getSimpleName()
								+ " failed: timeout");
					} catch (final ExecutionException e) {
						failedChecks++;
						out.println("\t - "
								+ checker.getClass().getSimpleName()
								+ " failed: " + e.getMessage());
					}
				}
				out.print("\t*******");
				for (int i = 0; i < Utils.getQualifiedName(method).length(); i++)
					out.print("*");
				out.println("*******");
			}
			out.print("*******");
			for (int i = 0; i < clazz.getName().length(); i++)
				out.print("*");
			out.println("*******");
		}
		out.println("\nRun Checks: " + runChecks + ", Failed: " + failedChecks);
		System.setOut(out);
	}

	/**
	 * Gets all the applicable class checkers for a given class.
	 * 
	 * @param clazz
	 *            the class to check
	 * @return the set of applicable class checkers
	 */
	public Set<ClassChecker> getCheckers(final Class<?> clazz) {
		final Set<ClassChecker> checkers = new HashSet<ClassChecker>();
		if (getProvider().canProvide(clazz))
			for (final Class<? extends ClassChecker> checkerClass : getClassCheckers()) {
				try {
					final ClassChecker checker = checkerClass.getConstructor(
							IInstanceProvider.class, Class.class).newInstance(
							getProvider(), clazz);
					if (checker.shouldCheck())
						checkers.add(checker);
				} catch (final IllegalArgumentException e) {
					LOGGER.error(
							"Error while instantiating {}, illegal argument exception: {}",
							checkerClass, e.getMessage());
				} catch (final SecurityException e) {
					LOGGER.error(
							"Error while instantiating {}, security exception: {}",
							checkerClass, e.getMessage());
				} catch (final InstantiationException e) {
					LOGGER.error(
							"Error while instantiating {}, instantiation exception: {}",
							checkerClass, e.getMessage());
				} catch (final IllegalAccessException e) {
					LOGGER.error(
							"Error while instantiating {}, illegal access exception: {}",
							checkerClass, e.getMessage());
				} catch (final InvocationTargetException e) {
					LOGGER.error(
							"Error while instantiating {}, invocation target exception: {}",
							checkerClass, e.getMessage());
				} catch (final NoSuchMethodException e) {
					throw new EasycoverageAssertionError(
							checkerClass.getCanonicalName()
									+ " should have a public constructor with two arguments (IInstanceProvider, Class)");
				}
			}
		return checkers;
	}

	/**
	 * Gets all the applicable method checkers for a given method.
	 * 
	 * @param method
	 *            the method to check
	 * @return the set of applicable method checkers
	 */
	public Set<MethodChecker> getCheckers(final Method method) {
		final Set<MethodChecker> checkers = new HashSet<MethodChecker>();
		if (method != null
				&& getProvider().canProvide(method.getDeclaringClass()))
			for (final Class<? extends MethodChecker> checkerClass : getMethodCheckers()) {
				try {
					final MethodChecker checker = checkerClass.getConstructor(
							IInstanceProvider.class, Method.class).newInstance(
							getProvider(), method);
					if (checker.shouldCheck())
						checkers.add(checker);
				} catch (final IllegalArgumentException e) {
					LOGGER.error(
							"Error while instantiating {}, illegal argument exception: {}",
							checkerClass, e.getMessage());
				} catch (final SecurityException e) {
					LOGGER.error(
							"Error while instantiating {}, security exception: {}",
							checkerClass, e.getMessage());
				} catch (final InstantiationException e) {
					LOGGER.error(
							"Error while instantiating {}, instantiation exception: {}",
							checkerClass, e.getMessage());
				} catch (final IllegalAccessException e) {
					LOGGER.error(
							"Error while instantiating {}, illegal access exception: {}",
							checkerClass, e.getMessage());
				} catch (final InvocationTargetException e) {
					LOGGER.error(
							"Error while instantiating {}, invocation target exception: {}",
							checkerClass, e.getMessage());
				} catch (final NoSuchMethodException e) {
					throw new EasycoverageAssertionError(
							checkerClass.getCanonicalName()
									+ " should have a public constructor with two arguments (IInstanceProvider, Class)");
				}
			}
		return checkers;
	}

	public IInstanceProvider getProvider() {
		if (provider == null)
			provider = new DefaultInstanceProvider();
		return provider;
	}

	public void setProvider(final IInstanceProvider provider) {
		this.provider = provider;
	}

	public Set<Class<?>> getClasses() {
		if (classes == null)
			classes = new HashSet<Class<?>>();
		return classes;
	}

	public void setClasses(final Set<Class<?>> classes) {
		this.classes = classes;
	}

	public Set<Class<? extends ClassChecker>> getClassCheckers() {
		if (classCheckers == null)
			classCheckers = new HashSet<Class<? extends ClassChecker>>();
		return classCheckers;
	}

	public void setClassCheckers(
			final Set<Class<? extends ClassChecker>> classCheckers) {
		this.classCheckers = classCheckers;
	}

	public Set<Class<? extends MethodChecker>> getMethodCheckers() {
		if (methodCheckers == null)
			methodCheckers = new HashSet<Class<? extends MethodChecker>>();
		return methodCheckers;
	}

	public void setMethodCheckers(
			final Set<Class<? extends MethodChecker>> methodCheckers) {
		this.methodCheckers = methodCheckers;
	}

	public int getTimeout() {
		return timeout;
	}

	public void setTimeout(final int timeout) {
		this.timeout = timeout;
	}

	private static class CheckerTask implements Runnable {

		private final IEasycoverageChecker	checker;

		public CheckerTask(final IEasycoverageChecker checker) {
			this.checker = checker;
		}

		public void run() {
			checker.check();
		}

	}

}
