package com.tocea.easycoverage.framework.checkers;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.tocea.easycoverage.framework.ClassChecker;
import com.tocea.easycoverage.framework.api.IInstanceProvider;
import com.tocea.easycoverage.framework.exception.EasycoverageAssertionError;
import com.tocea.easycoverage.framework.utils.Utils;

/**
 * 
 * Checkers that verifies that, for each class implementing Cloneable, that
 * o.equals(o.clone).
 * 
 * @author Romain Delamare
 * 
 */
public class CloneChecker extends ClassChecker {

	private static final Logger	LOGGER	= LoggerFactory
												.getLogger(CloneChecker.class);

	public CloneChecker(final IInstanceProvider provider, final Class<?> clazz) {
		super(provider, clazz);
	}

	public void check() {
		if (provider == null || clazz == null)
			return;
		final Object t = provider.getInstance(clazz);
		try {
			final Method clone = clazz.getMethod("clone");
			clone.setAccessible(true);
			final Object o = clone.invoke(t);
			if (!t.equals(o))
				throw new EasycoverageAssertionError(
						"Clone error: o.equals(o.clone()) should always be true.");
		} catch (final SecurityException e) {
			LOGGER.error(
					"Error : security exception while invoking clone() on {}",
					clazz.getCanonicalName());
		} catch (final IllegalArgumentException e) {
			LOGGER.error(
					"Error : illegal argument exception while invoking clone() on {}",
					clazz.getCanonicalName());
		} catch (final IllegalAccessException e) {
			LOGGER.error(
					"Error : illegal access exception while invoking clone() on {}",
					clazz.getCanonicalName());
		} catch (final NoSuchMethodException e) {
			throw new EasycoverageAssertionError(
					"The "
							+ clazz.getCanonicalName()
							+ " class implements Cloneable but does not implement clone().");
		} catch (final InvocationTargetException e) {
			throw new RuntimeException(e.getCause());
		}
	}

	@Override
	public boolean shouldCheck() {
		final List<Class<?>> interfaces = new ArrayList<Class<?>>();
		Utils.getAllInterfaces(clazz, interfaces);
		return super.shouldCheck() && interfaces.contains(Cloneable.class);
	}

}
