package com.tocea.easycoverage.framework.providers;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.tocea.easycoverage.framework.api.IInstanceProvider;

/**
 * 
 * Instance provider based on the copy constructor pattern. It can provide a new
 * instance for any class that has an associated reference instance. The
 * reference instance is copied using a copy constructor. A copy constructor
 * takes an instance of the same class as only parameter and copies attributes.
 * The type of copy (shallow or deep) is left to the copy constructor.
 * 
 * @author Romain Delamare
 * 
 */
public class CopyConstructorInstanceProvider implements IInstanceProvider {

	private static final Logger			LOGGER		= LoggerFactory
															.getLogger(CopyConstructorInstanceProvider.class);

	private final Map<Class<?>, Object>	instances	= new HashMap<Class<?>, Object>();

	/**
	 * Adds a reference instance. It is associated only with its actual class.
	 * 
	 * @param instance
	 *            the new reference instance
	 */
	public void addInstance(final Object instance) {
		if (instance != null && getConstructor(instance.getClass()) != null)
			instances.put(instance.getClass(), instance);
	}

	/**
	 * Associates a reference instance to a class. The instance could actually
	 * be a subtype of the class for which it is set.
	 * 
	 * @param clazz
	 *            the class for which the instance is set
	 * @param instance
	 *            the instance to set
	 */
	public <T> void setInstance(final Class<T> clazz, final T instance) {
		if (instance != null && getConstructor(clazz) != null)
			instances.put(clazz, instance);
	}

	public <T> T getInstance(final Class<T> clazz) {
		T result = null;
		if (clazz != null)
			try {
				result = getConstructor(clazz)
						.newInstance(instances.get(clazz));
			} catch (final IllegalArgumentException e) {
				LOGGER.error(
						"Error instantiating {}, illegal argument exception: {}",
						clazz.getCanonicalName(), e.getMessage());
			} catch (final InstantiationException e) {
				LOGGER.error(
						"Error instantiating {}, instantiation exception: {}",
						clazz.getCanonicalName(), e.getMessage());
			} catch (final IllegalAccessException e) {
				LOGGER.error(
						"Error instantiating {}, illegal access exception: {}",
						clazz.getCanonicalName(), e.getMessage());
			} catch (final InvocationTargetException e) {
				LOGGER.error(
						"Error instantiating {}, invocation target exception: {}",
						clazz.getCanonicalName(), e.getMessage());
			}
		return result;
	}

	public <T> boolean canProvide(final Class<T> clazz) {
		return instances.containsKey(clazz);
	}

	private static <T> Constructor<T> getConstructor(final Class<T> clazz) {
		Constructor<T> constructor = null;
		try {
			constructor = clazz.getConstructor(clazz);
		} catch (final SecurityException e) {
			// Nothing to do: we cannot create an instance
		} catch (final NoSuchMethodException e) {
			// Nothing to do we cannot create an instance
		}
		return constructor;
	}

}
