package com.tocea.easycoverage.framework.utils;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * 
 * Utility class for EasyCoverage and its checkers.
 * 
 * @author Romain Delamare
 * 
 */
public class Utils {

	/**
	 * @param method
	 *            a method
	 * @return the fully qualified name of the method in parameter
	 */
	public static String getQualifiedName(final Method method) {
		if (method == null)
			return "";
		final StringBuilder sb = new StringBuilder();
		sb.append(method.getDeclaringClass().getName());
		sb.append('.');
		sb.append(method.getName());
		sb.append('(');
		final Class<?>[] params = method.getParameterTypes();
		for (int i = 0; i < params.length; i++) {
			sb.append(getFieldName(params[i]));
			if (i < (params.length - 1))
				sb.append(",");
		}
		sb.append(')');
		return sb.toString();
	}

	public static String getFieldName(final Class<?> type) {
		if (type == null)
			return "";
		if (type.isArray()) {
			try {
				Class<?> cl = type;
				int dimensions = 0;
				while (cl.isArray()) {
					dimensions++;
					cl = cl.getComponentType();
				}
				final StringBuilder sb = new StringBuilder(cl.getName());
				for (int i = 0; i < dimensions; i++) {
					sb.append("[]");
				}
				return sb.toString();
			} catch (final Throwable e) {
				/* FALLTHRU */
			}
		}
		return type.getName();
	}

	/**
	 * Generates default parameter for a method or constructor. Objects are set
	 * to null, primitive types are set to a default value.
	 * 
	 * @param parametersType
	 *            an array of classes
	 * @return an array of instances, with types corresponding to the ones in
	 *         parametersType
	 */
	public static Object[] generateDefaultValuesParameters(
			final Class<?>[] parametersType) {
		if (parametersType == null)
			return null;
		final Object[] parameters = new Object[parametersType.length];
		for (int i = 0; i < parametersType.length; i++) {
			if (int.class.equals(parametersType[i])
					|| Integer.class.equals(parametersType[i])
					|| short.class.equals(parametersType[i])
					|| Short.class.equals(parametersType[i])
					|| byte.class.equals(parametersType[i])
					|| Byte.class.equals(parametersType[i])
					|| long.class.equals(parametersType[i])
					|| Long.class.equals(parametersType[i])) {
				parameters[i] = 0;
			} else if (float.class.equals(parametersType[i])
					|| Float.class.equals(parametersType[i])) {
				parameters[i] = 0f;
			} else if (double.class.equals(parametersType[i])
					|| Double.class.equals(parametersType[i])) {
				parameters[i] = 0d;
			} else if (char.class.equals(parametersType[i])
					|| Character.class.equals(parametersType[i])) {
				parameters[i] = '\u0000';
			} else if (boolean.class.equals(parametersType[i])
					|| Boolean.class.equals(parametersType[i])) {
				parameters[i] = false;
			} else {
				parameters[i] = null;
			}
		}
		return parameters;
	}

	/**
	 * Generate parameters for a method or constructor, with max values. Objects
	 * are set to null.
	 * 
	 * @param parametersType
	 *            an array of classes
	 * @return an array of instances, with types corresponding to the ones in
	 *         parametersType
	 */
	public static Object[] generateMaxValues(final Class<?>[] parametersType) {
		if (parametersType == null)
			return null;
		final Object[] parameters = new Object[parametersType.length];
		for (int i = 0; i < parametersType.length; i++) {
			if (int.class.equals(parametersType[i])
					|| Integer.class.equals(parametersType[i])) {
				parameters[0] = Integer.MAX_VALUE;
			} else if (short.class.equals(parametersType[i])
					|| Short.class.equals(parametersType[i])) {
				parameters[0] = Short.MAX_VALUE;
			} else if (byte.class.equals(parametersType[i])
					|| Byte.class.equals(parametersType[i])) {
				parameters[0] = Byte.MAX_VALUE;
			} else if (long.class.equals(parametersType[i])
					|| Long.class.equals(parametersType[i])) {
				parameters[i] = Long.MAX_VALUE;
			} else if (float.class.equals(parametersType[i])
					|| Float.class.equals(parametersType[i])) {
				parameters[i] = Float.MAX_VALUE;
			} else if (double.class.equals(parametersType[i])
					|| Double.class.equals(parametersType[i])) {
				parameters[i] = Double.MAX_VALUE;
			} else if (char.class.equals(parametersType[i])
					|| Character.class.equals(parametersType[i])) {
				parameters[i] = '\u0000';
			} else if (boolean.class.equals(parametersType[i])
					|| Boolean.class.equals(parametersType[i])) {
				parameters[i] = false;
			} else {
				parameters[i] = null;
			}
		}
		return parameters;
	}

	/**
	 * Generate parameters for a method or constructor, with min values. Objects
	 * are set to null.
	 * 
	 * @param parametersType
	 *            an array of classes
	 * @return an array of instances, with types corresponding to the ones in
	 *         parametersType
	 */
	public static Object[] generateMinValues(final Class<?>[] parametersType) {
		if (parametersType == null)
			return null;
		final Object[] parameters = new Object[parametersType.length];
		for (int i = 0; i < parametersType.length; i++) {
			if (int.class.equals(parametersType[i])
					|| Integer.class.equals(parametersType[i])) {
				parameters[0] = Integer.MIN_VALUE;
			} else if (short.class.equals(parametersType[i])
					|| Short.class.equals(parametersType[i])) {
				parameters[0] = Short.MIN_VALUE;
			} else if (byte.class.equals(parametersType[i])
					|| Byte.class.equals(parametersType[i])) {
				parameters[0] = Byte.MIN_VALUE;
			} else if (long.class.equals(parametersType[i])
					|| Long.class.equals(parametersType[i])) {
				parameters[i] = Long.MIN_VALUE;
			} else if (float.class.equals(parametersType[i])
					|| Float.class.equals(parametersType[i])) {
				parameters[i] = Float.MIN_VALUE;
			} else if (double.class.equals(parametersType[i])
					|| Double.class.equals(parametersType[i])) {
				parameters[i] = Double.MIN_VALUE;
			} else if (char.class.equals(parametersType[i])
					|| Character.class.equals(parametersType[i])) {
				parameters[i] = '\u0000';
			} else if (boolean.class.equals(parametersType[i])
					|| Boolean.class.equals(parametersType[i])) {
				parameters[i] = false;
			} else {
				parameters[i] = null;
			}
		}
		return parameters;
	}

	/**
	 * Computes all the interfaces implemented by a given class, transitively.
	 * 
	 * @param _class
	 *            a class
	 * @param interfaces
	 *            all the interfaces directly implemented by _class or one of
	 *            its parents
	 */
	public static void getAllInterfaces(final Class<?> _class,
			final List<Class<?>> interfaces) {
		if (_class != null && interfaces != null) {
			interfaces.addAll(Arrays.asList(_class.getInterfaces()));
			if (_class.getSuperclass() != null) {
				getAllInterfaces(_class.getSuperclass(), interfaces);
			}
		}
	}

	/**
	 * Checks if a method is a getter.
	 * 
	 * @param method
	 *            a method
	 * @return true if method is a getter, false otherwise
	 */
	public static boolean isGetter(final Method method) {
		return method != null
				&& method.getParameterTypes().length == 0
				&& (method.getName().startsWith("get") || ((method.getName()
						.startsWith("is") || method.getName().startsWith("has")) && (method
						.getReturnType() == Boolean.class || method
						.getReturnType() == boolean.class)))
				&& method.getReturnType() != void.class;
	}

	/**
	 * Checks if a method is a setter.
	 * 
	 * @param method
	 *            a method
	 * @return true if method is a setter, false otherwise
	 */
	public static boolean isSetter(final Method method) {
		return method != null && method.getParameterTypes().length == 1
				&& method.getReturnType() == void.class
				&& method.getName().startsWith("set");
	}

	/**
	 * Computes all the public methods of a class, included the methods declared
	 * by one of its parents.
	 * 
	 * @param clazz
	 *            class
	 * @return all the public method declared for clazz
	 */
	public static Set<Method> getDeclaredPublicMethods(final Class<?> clazz) {
		final Set<Method> methods = new HashSet<Method>();
		if (clazz != null)
			for (final Method method : clazz.getDeclaredMethods())
				if (Modifier.isPublic(method.getModifiers())
						&& !method.isBridge() && !method.isSynthetic())
					methods.add(method);
		return methods;
	}

	/**
	 * Returns the setter corresponding to a getter
	 * 
	 * @param method
	 *            a getter method
	 * @return the setter corresponding to the given getter, null if there is
	 *         none
	 */
	public static Method getGetterForSetter(final Method method) {
		if (method != null) {
			final Class<?> _class = method.getDeclaringClass();
			final String name = method.getName().substring(3);
			for (final Method m : getDeclaredPublicMethods(_class)) {
				if (isGetter(m) && name.equals(m.getName().substring(3))) {
					return m;
				}
			}
		}
		return null;
	}

}
